/*
 *  chardev.c -     
 *                /
 */

#include <linux/module.h> /*     */
#include <linux/kernel.h> /* -    ! */
#include <linux/fs.h>
#include <asm/uaccess.h>	/*   get_user  put_user */

#include "chardev.h"
#define SUCCESS 0
#define DEVICE_NAME "char_dev"
#define BUF_LEN 80

/* 
 *   ?  
 *     
 */
static int Device_Open = 0;

/* 
 *    
 */
static char Message[BUF_LEN];

/* 
 *   .
 *    ,    
 *   . 
 */
static char *Message_Ptr;

/* 
 *       
 */
static int device_open(struct inode *inode, struct file *file)
{
#ifdef DEBUG
  printk("device_open(%p)\n", file);
#endif

  /* 
   *             
   */
  if (Device_Open)
    return -EBUSY;

  Device_Open++;
  /*
   *  
   */
  Message_Ptr = Message;
  try_module_get(THIS_MODULE);
  return SUCCESS;
}

static int device_release(struct inode *inode, struct file *file)
{
#ifdef DEBUG
  printk("device_release(%p,%p)\n", inode, file);
#endif

  /* 
   *        
   */
  Device_Open--;

  module_put(THIS_MODULE);
  return SUCCESS;
}

/* 
 *   ,   
 *     .
 */
static ssize_t device_read(struct file *file,	/* . include/linux/fs.h   */
            char __user * buffer,	/*    */
            size_t length,        /*         */
            loff_t * offset)
{
  /* 
   *  ,    
   */
  int bytes_read = 0;

#ifdef DEBUG
  printk("device_read(%p,%p,%d)\n", file, buffer, length);
#endif

  /* 
   *     --  0
   * (  ) 
   */
  if (*Message_Ptr == 0)
    return 0;

  /* 
   *     
   */
  while (length && *Message_Ptr) {

    /* 
     *      ,
     *    . 
     *     put_user,
     *      
     *   .
     */
    put_user(*(Message_Ptr++), buffer++);
    length--;
    bytes_read++;
  }

#ifdef DEBUG
  printk("Read %d bytes, %d left\n", bytes_read, length);
#endif

  /* 
   *   ,   .
   */
	return bytes_read;
}

/* 
 *       
 */
static ssize_t
device_write(struct file *file,
            const char __user * buffer, size_t length, loff_t * offset)
{
  int i;

#ifdef DEBUG
  printk("device_write(%p,%s,%d)", file, buffer, length);
#endif

  for (i = 0; i < length && i < BUF_LEN; i++)
    get_user(Message[i], buffer + i);

  Message_Ptr = Message;

  /* 
   *    
   */
  return i;
}

/* 
 * ,      ioctl   .
 *  inode   file     :
 *  ioctl   .
 *
 *   ioctl     ,
 *       .
 *
 */
int device_ioctl(struct inode *inode,	/* . include/linux/fs.h */
      struct file *file,              /*    */
      unsigned int ioctl_num,         /*    ioctl */
      unsigned long ioctl_param)
{
  int i;
  char *temp;
  char ch;

  /* 
   *     ioctl 
   */
  switch (ioctl_num) {
  case IOCTL_SET_MSG:
    /* 
     *     (  ) 
     *    .      .
     */
    temp = (char *)ioctl_param;

    /* 
     *   
     */
    get_user(ch, temp);
    for (i = 0; ch && i < BUF_LEN; i++, temp++)
      get_user(ch, temp);

    device_write(file, (char *)ioctl_param, i, 0);
    break;

  case IOCTL_GET_MSG:
    /* 
     *      - 
     *    .
     */
    i = device_read(file, (char *)ioctl_param, 99, 0);

    /* 
     *      \0
     */
    put_user('\0', (char *)ioctl_param + i);
    break;

  case IOCTL_GET_NTH_BYTE:
    /* 
     *     (ioctl_param)  
     *  (  ) 
     */
    return Message[ioctl_param];
    break;
  }

  return SUCCESS;
}

/*  */

/* 
 *      -
 * ,    .
 *         ,
 *       
 * init_module.       NULL.
 */
struct file_operations Fops = {
  .read = device_read,
  .write = device_write,
  .ioctl = device_ioctl,
  .open = device_open,
  .release = device_release,	/*   close */
};

/* 
 *   -   
 */
int init_module()
{
  int ret_val;
  /* 
   *    (   -  ) 
   */
  ret_val = register_chrdev(MAJOR_NUM, DEVICE_NAME, &Fops);

  /* 
   *    
   */
  if (ret_val < 0) {
    printk("%s failed with %d\n",
           "Sorry, registering the character device ", ret_val);
    return ret_val;
  }

  printk("%s The major device number is %d.\n",
         "Registeration is a success", MAJOR_NUM);
  printk("If you want to talk to the device driver,\n");
  printk("you'll have to create a device file. \n");
  printk("We suggest you use:\n");
  printk("mknod %s c %d 0\n", DEVICE_FILE_NAME, MAJOR_NUM);
  printk("The device file name is important, because\n");
  printk("the ioctl program assumes that's the\n");
  printk("file you'll use.\n");

  return 0;
}

/* 
 *    -    /proc 
 */
void cleanup_module()
{
  int ret;

  /* 
   *  
   */
  ret = unregister_chrdev(MAJOR_NUM, DEVICE_NAME);

  /* 
   *    --  
   */
  if (ret < 0)
    printk("Error in module_unregister_chrdev: %d\n", ret);
}
