/*
 *  sleep.c -    /proc,        ,
 *     .
 */

#include <linux/kernel.h>   /* -    ! */
#include <linux/module.h>   /*     */
#include <linux/proc_fs.h>	/*     /proc */
#include <linux/sched.h>    /*    */
#include <asm/uaccess.h>    /*   get_user  put_user */

/* 
 *     ,
 *     ,  , 
 *       
 */
#define MESSAGE_LENGTH 80
static char Message[MESSAGE_LENGTH];

static struct proc_dir_entry *Our_Proc_File;
#define PROC_ENTRY_FILENAME "sleep"

static ssize_t module_output(struct file *file, /* . include/linux/fs.h   */
                      char *buf,         /*    (  ) */
                      size_t len,	      /*   */
                      loff_t * offset)
{
  static int finished = 0;
  int i;
  char message[MESSAGE_LENGTH + 30];

  /* 
   *       0.
   *         
   *    . 
   */
	if (finished) {
		finished = 0;
		return 0;
	}

  /* 
   *          
   *   put_user.
   *    -- get_user.
   */
  sprintf(message, "Last input:%s\n", Message);
  for (i = 0; i < len && message[i]; i++)
    put_user(message[i], buf + i);

  finished = 1;
  return i;    /*   ""  */
}

/* 
 *      
 */
static ssize_t module_input(struct file *file,  /*   */
          const char *buf,                      /*    */
          size_t length,                        /*   */
          loff_t * offset)
{         /*    -  */
  int i;

  /* 
   *  ,     , 
   *      module_output.
   */
  for (i = 0; i < MESSAGE_LENGTH - 1 && i < length; i++)
    get_user(Message[i], buf + i);

  /*  ,   \0 */    	
  Message[i] = '\0';

  /* 
   *    
   */
  return i;
}

/* 
 * 1 --   
 */
int Already_Open = 0;

/* 
 *  
 */
DECLARE_WAIT_QUEUE_HEAD(WaitQ);
/* 
 *      /proc 
 */
static int module_open(struct inode *inode, struct file *file)
{
  /* 
   *    O_NONBLOCK,     
   *   ,    ,    
   * -EAGAIN,   "   "
   */
  if ((file->f_flags & O_NONBLOCK) && Already_Open)
    return -EAGAIN;

  /* 
   *   ,
   *     
   */
  try_module_get(THIS_MODULE);

  /* 
   *     --  
   */

  while (Already_Open) {
    int i, is_sig = 0;

    /* 
     *          .
     *      ,   
     *  ,      
     * wake_up(&WaitQ) (    module_close,  
     *   )      Ctrl-C
     */
    wait_event_interruptible(WaitQ, !Already_Open);

    for (i = 0; i < _NSIG_WORDS && !is_sig; i++)
    is_sig =
          current->pending.signal.sig[i] & ~current->
          blocked.sig[i];

    if (is_sig) {
      /* 
       *     module_put(THIS_MODULE),
       *    
       *      close.
       *     ,     
       *  ,     
       *     
       */
      module_put(THIS_MODULE);
      return -EINTR;
    }
  }

  /* 
   *     Already_Open    
   */

  /* 
   *  
   */
  Already_Open = 1;
  return 0;
}

/* 
 *    
 */
int module_close(struct inode *inode, struct file *file)
{
  /* 
   *    Already_Open,      WaitQ
   *       .
   *   ,       
   */
  Already_Open = 0;

  /* 
   *     WaitQ.
   */
  wake_up(&WaitQ);

  module_put(THIS_MODULE);

  return 0;
}

/* 
 *           
 * 0 -- ,   -- .
 *
 *     :
 * 0 - e (     )
 * 2 -  (     )
 * 4 -  (     )
 *
 *       
 * ,   ls -l 
 *    .
 */
static int module_permission(struct inode *inode, int op, struct nameidata *nd)
{
  /* 
   *    , 
   *  --  root- (uid 0)
   */
  if (op == 4 || (op == 2 && current->euid == 0))
    return 0;

  /* 
   *  -  --  
   */
  return -EACCES;
}

/* 
 *   -   .
 */
static struct file_operations File_Ops_4_Our_Proc_File = {
  .read = module_output,    /*    */
  .write = module_input,    /*    */
  .open = module_open,      /*   */
  .release = module_close,  /*   */
};


/* 
 *      . 
 *  ,    
 * file_operations  ,   ,  
 *      .    
 *  -,     .
 */
static struct inode_operations Inode_Ops_4_Our_Proc_File = {
  .permission = module_permission,	/* check for permissions */
};

/* 
 *     
 */

/* 
 *   -    /proc
 */

int init_module()
{
  int rv = 0;
  Our_Proc_File = create_proc_entry(PROC_ENTRY_FILENAME, 0644, NULL);
  Our_Proc_File->owner = THIS_MODULE;
  Our_Proc_File->proc_iops = &Inode_Ops_4_Our_Proc_File;
  Our_Proc_File->proc_fops = &File_Ops_4_Our_Proc_File;
  Our_Proc_File->mode = S_IFREG | S_IRUGO | S_IWUSR;
  Our_Proc_File->uid = 0;
  Our_Proc_File->gid = 0;
  Our_Proc_File->size = 80;

  if (Our_Proc_File == NULL) {
    rv = -ENOMEM;
    remove_proc_entry(PROC_ENTRY_FILENAME, &proc_root);
    printk(KERN_INFO "Error: Could not initialize /proc/test\n");
  }

  return rv;
}

/* 
 *    -    /proc.  
 *   WaitQ  ,   ,    
 *     open,      . 
 * ,  9 ,         
 */
void cleanup_module()
{
  remove_proc_entry(PROC_ENTRY_FILENAME, &proc_root);
}
