/* 
 *  print_string.c -    tty ,    
 *  X11 ,  telnet,  - .       tty,
 *     .
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>   /*  current */
#include <linux/tty.h>     /*  tty */
#include <linux/version.h> /*  LINUX_VERSION_CODE */

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Peter Jay Salzman");

static void print_string(char *str)
{
  struct tty_struct *my_tty;

  /* 
   *    2.6.6,  tty    signal 
   */
#if ( LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,5) )
  /* 
   * tty  
   */
  my_tty = current->tty;
#else
  /* 
   * tty  ,   2.6.6  
   */
  my_tty = current->signal->tty;
#endif

  /* 
   *  my_tty == NULL,      tty    
   *      (, ).  
   *        .
   */
  if (my_tty != NULL) {

    /* 
     * my_tty->driver -- ,     -,
     *    (write)      tty. 
     *
     *    -- tty,    ,
     *          
     * tty     . 
     *   --   
     *      ,   false (0)
     *    ,  true ( ).  
     *   --   .
     *   --  .
     */
    ((my_tty->driver)->write) (my_tty,  /*  tty  */
             0,                         /*     */
             str,                       /*   */
             strlen(str));              /*   */

    /* 
     * tty    ,  () 
     *   ASCII,    
     *      -- " "  " ".
     *  Unix,  ASCII -- " "    ,
     *       
     *  .
     *
     *         Unix 
     * MS Windows. CP/M   "",  MS-DOS 
     * MS Windows,    ASCII.
     */
    ((my_tty->driver)->write) (my_tty, 0, "\015\012", 2);
  }
}

static int __init print_string_init(void)
{
  print_string("The module has been inserted.  Hello world!");
  return 0;
}

static void __exit print_string_exit(void)
{
  print_string("The module has been removed.  Farewell world!");
}

module_init(print_string_init);
module_exit(print_string_exit);
